/*******************************************************************************
 * Copyright 2017 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europäischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie dürfen dieses Werk ausschließlich gemäß dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschränkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/
package de.frankmuenster.jameica.finanzen.boerseard;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.math.BigDecimal;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.net.ssl.HttpsURLConnection;

import org.apache.commons.lang.StringEscapeUtils;
import org.apache.commons.lang.StringUtils;

import de.frankmuenster.jameica.finanzen.AdditionalHtmlTags;
import de.frankmuenster.jameica.finanzen.DatenQuelleApi;
import de.frankmuenster.jameica.finanzen.FinanceException;
import de.frankmuenster.jameica.finanzen.FinanceExceptionType;
import de.frankmuenster.jameica.finanzen.StockInfoBean;
import de.frankmuenster.jameica.finanzen.WertpapierHistory;
import de.frankmuenster.jameica.finanzen.boerseard.json.JsonArdBoerseHistory;
import de.frankmuenster.jameica.json.JsonReader;
import de.frankmuenster.jameica.json.exeptions.JsonCastException;
import de.willuhn.logging.Level;
import de.willuhn.logging.Logger;

/**
 * @author Frank Münster
 */
@SuppressWarnings("PMD.ExcessiveImports")
public final class BoerseArdApi implements DatenQuelleApi<ArdStockInfoBean> {

    /**
     * Identifier of history data
     */
    private static final String US_FDATA_HISTORY = "USFdata.history"; //$NON-NLS-1$

    /**
     *
     */
    private static final String URL_FORMAT_ERROR = "URL ist nicht korrekt formatiert."; //$NON-NLS-1$

    /**
     * Unser Instanz
     */
    private static DatenQuelleApi<ArdStockInfoBean> instance;

    /**
     * Haupt-URL zur Website
     */
    private static final String MAIN_URL = "https://kurse.boerse.ard.de/"; //$NON-NLS-1$

    /**
     * Standardantwort, falls nichts gefunden wurde.
     */
    private static final String NOT_A_RESOURCE = "NaR"; //$NON-NLS-1$

    /**
     * Such URL
     */
    private static final String SEARCH_URL = BoerseArdApi.MAIN_URL
            + "ard/ajax/get_search_json_dropdown.htn?&searchString=XXX&jsonp_callback=callbackInfo"; //$NON-NLS-1$

    /**
     * Wertpapier-Detail-URL
     */
    private static final String DETAIL_URL = BoerseArdApi.MAIN_URL + "ard/kurse_einzelkurs_uebersicht.htn?i="; //$NON-NLS-1$

    /**
     * Liste der schon abgefragte URLs
     */
    private final ConcurrentHashMap<URI, String> cacheUrlResponse = new ConcurrentHashMap<>();

    /**
     * Liefert die BoerseArd-API Instanz. Wird bei bedarf erstellt.
     *
     * @return die BoerseArd-Instanz
     */
    public static DatenQuelleApi<ArdStockInfoBean> getInstance() {

        synchronized (BoerseArdApi.class) {
            if (BoerseArdApi.instance == null) {
                BoerseArdApi.instance = new BoerseArdApi();
            }
        }
        return BoerseArdApi.instance;
    }

    /**
     * Privater Konstruktor
     */
    private BoerseArdApi() {
        super();
    }

    /**
     * @see DatenQuelleApi#getActualValue(StockInfoBean)
     */
    @Override
    public ArdStockInfoBean getActualValue(final ArdStockInfoBean stock) throws FinanceException {

        if (StringUtils.isEmpty(stock.getId())) {
            throw new FinanceException("ID des Wertpapiers nicht gesetzt", FinanceExceptionType.DATA_NOT_FOUND); //$NON-NLS-1$
        }
        final String uriString = BoerseArdApi.DETAIL_URL + stock.getId();
        try {
            final URI uri = new URI(uriString);
            final String response = this.getResonseGETRequest(uri);
            if (response.contains("aktueller Wert")) { //$NON-NLS-1$ )
                this.updateValues(stock, response);
            } else {
                throw new FinanceException("Wertpapier nicht gefunden", FinanceExceptionType.DATA_NOT_FOUND); //$NON-NLS-1$
            }
        } catch (final NumberFormatException e) {
            Logger.error("Wert ist falsch formatiert. ", e); //$NON-NLS-1$
            throw new FinanceException("Wert ist falsch formatiert " //$NON-NLS-1$ //
                    + e.getMessage(), FinanceExceptionType.OTHER, e);
        } catch (final URISyntaxException e) {
            Logger.error(BoerseArdApi.URL_FORMAT_ERROR, e);
            throw (FinanceException) new FinanceException(BoerseArdApi.URL_FORMAT_ERROR + e.getMessage(),
                    FinanceExceptionType.NETWORK).initCause(e);
        }
        return stock;
    }

    /**
     * @see DatenQuelleApi# getSearchResults(java.lang.String)
     */
    @Override
    public List<ArdStockInfoBean> getSearchResults(final String searchText) throws FinanceException {

        return this.search(searchText);
    }

    /**
     * @see DatenQuelleApi#loadHistoryValues(de.frankmuenster.jameica.finanzen.StockInfoBean)
     */
    @Override
    public ArdStockInfoBean loadHistoryValues(final ArdStockInfoBean stock) throws FinanceException {

        if (StringUtils.isEmpty(stock.getId())) {
            throw new FinanceException("ID des Wertpapiers nicht gesetzt", FinanceExceptionType.DATA_NOT_FOUND); //$NON-NLS-1$
        }
        final String uriString = BoerseArdApi.DETAIL_URL + stock.getId();
        try {
            final URI uri = new URI(uriString);
            final String response = this.getResonseGETRequest(uri, true);
            if (response.contains(BoerseArdApi.US_FDATA_HISTORY)) {
                this.readHistory(stock, response);
            } else {
                throw new FinanceException("Wertpapier nicht gefunden", FinanceExceptionType.DATA_NOT_FOUND); //$NON-NLS-1$
            }
        } catch (final URISyntaxException e) {
            Logger.error(BoerseArdApi.URL_FORMAT_ERROR, e);
            throw (FinanceException) new FinanceException(BoerseArdApi.URL_FORMAT_ERROR + e.getMessage(),
                    FinanceExceptionType.NETWORK).initCause(e);
        } catch (final NumberFormatException e) {
            Logger.error("Wert ist falsch formatiert.", e); //$NON-NLS-1$
            throw new FinanceException("Wert ist falsch formatiert " //$NON-NLS-1$
                    + e.getMessage(), FinanceExceptionType.OTHER, e);
        } catch (final JsonCastException e) {
            Logger.error("SON Daten konnten nicht geladen werden.", e); //$NON-NLS-1$
            throw new FinanceException("JSON Daten konnten nicht geladen werden. " //$NON-NLS-1$
                    + e.getMessage(), FinanceExceptionType.OTHER, e);
        }
        return stock;
    }

    private ConcurrentHashMap<String, String> getExchanges(final String content) {

        final ConcurrentHashMap<String, String> list = new ConcurrentHashMap<>();
        int startPos = content.indexOf("<h2>B&ouml;rsenpl&auml;tze"); //$NON-NLS-1$
        if (startPos > 0) {
            startPos = content.indexOf(AdditionalHtmlTags.TAG_TBODY_BEGIN, startPos);
            final int endTbody = content.indexOf(AdditionalHtmlTags.TAG_TBODY_END, startPos);
            final String tbody = content.substring(startPos, endTbody);
            int startRow = tbody.indexOf(AdditionalHtmlTags.TAG_TR_BEGIN);
            while (startRow > 0) {
                // Lese ersten Link
                final int startLink = tbody.indexOf(AdditionalHtmlTags.TAG_A_BEGIN, startRow);
                final int endLink = tbody.indexOf(AdditionalHtmlTags.TAG_CLOSE, startLink);
                final int startHref = tbody.indexOf("href=", startLink) + 6; //$NON-NLS-1$
                final int endHref = tbody.indexOf('\"', startHref);
                final int startEndLinkTag = tbody.indexOf(AdditionalHtmlTags.TAG_A_END, endLink);
                final String exchangeName = tbody.substring(endLink + 1, startEndLinkTag);
                String exchangeLink = tbody.substring(startHref, endHref);
                final int paramStart = exchangeLink.indexOf("i="); //$NON-NLS-1$
                exchangeLink = exchangeLink.substring(paramStart + 2);
                Logger.debug("Börse : " + exchangeName + ": " + exchangeLink); //$NON-NLS-1$ //$NON-NLS-2$
                list.put(exchangeName, exchangeLink);
                startRow = tbody.indexOf(AdditionalHtmlTags.TAG_TR_BEGIN, endLink);
            }
        }
        return list;
    }

    /**
     * Liefert die Antwort eines Web-Services Aufrufs.
     *
     * @param uri
     *            des Webservices.
     * @return die Antwort des Web Service, oder <code>"NaR"</code> falls mit
     *         dem Aufruf alles in Ordnung war, aber keine Antwort kam.
     * @throws FinanceException
     *             falls w&auml;hrend des Aufrufs etwas schief ging.
     */
    private String getResonseGETRequest(final URI uri) throws FinanceException {

        return this.getResonseGETRequest(uri, false);
    }

    /**
     * Liefert die Antwort eines Web-Services Aufrufs.
     *
     * @param uri
     *            des Webservices.
     * @param useCache
     *            <code>true</code> falls der interne Cache verwendet werden
     *            soll, sonst <code>false</code>.
     * @return die Antwort des Web Service, oder <code>"NaR"</code> falls mit
     *         dem Aufruf alles in Ordnung war, aber keine Antwort kam.
     * @throws FinanceException
     *             falls w&auml;hrend des Aufrufs etwas schief ging.
     */
    private String getResonseGETRequest(final URI uri, final boolean useCache) throws FinanceException {

        String response;
        if (useCache && this.cacheUrlResponse.containsKey(uri)) {
            Logger.debug("Lese aus Cache"); //$NON-NLS-1$
            response = this.cacheUrlResponse.get(uri);
        } else {
            response = BoerseArdApi.NOT_A_RESOURCE;
            try {
                if (Level.DEBUG.equals(Logger.getLevel())) {
                    Logger.debug("URL: " + uri.toString()); //$NON-NLS-1$
                }
                final HttpsURLConnection conn = (HttpsURLConnection) uri.toURL().openConnection();
                conn.setRequestMethod("GET"); //$NON-NLS-1$
                conn.setConnectTimeout(2000);
                conn.setRequestProperty("Accept-Charset", StandardCharsets.UTF_8.name()); //$NON-NLS-1$
                conn.setRequestProperty("User-Agent", "JaDepot"); //$NON-NLS-1$ //$NON-NLS-2$
                Logger.debug(conn.getResponseMessage());
                if (conn.getResponseCode() != HttpURLConnection.HTTP_OK) {
                    Logger.error("Repose Error: " + conn.getResponseCode()); //$NON-NLS-1$
                    throw new FinanceException(
                            "Netzwerkfehler aufgetreten. HTTP Fehlercode:" + Integer.toString(conn.getResponseCode()), //$NON-NLS-1$
                            FinanceExceptionType.NETWORK);
                }
                final BufferedReader bufferReader = new BufferedReader(
                        new InputStreamReader(conn.getInputStream(), StandardCharsets.UTF_8.name()));
                final StringBuilder sBuffer = new StringBuilder(StringUtils.EMPTY);
                String output;
                while ((output = bufferReader.readLine()) != null) { // NOPMD
                    sBuffer.append(output).append(System.lineSeparator());
                }
                response = sBuffer.toString();
                conn.disconnect();
                if (useCache) {
                    this.cacheUrlResponse.put(uri, response);
                }
            } catch (final IOException e) {
                Logger.debug(e.getLocalizedMessage());
                throw (FinanceException) new FinanceException("IOException: " + e.getMessage(), //$NON-NLS-1$
                        FinanceExceptionType.NETWORK).initCause(e);
            }
        }
        return response;
    }

    /**
     * @param stock
     * @param response
     * @throws JsonCastException
     */
    private void readHistory(final ArdStockInfoBean stock, final String response) throws JsonCastException {

        int start = response.indexOf(BoerseArdApi.US_FDATA_HISTORY) + 17;
        int end = response.indexOf("});", start) + 1; //$NON-NLS-1$
        String json = response.substring(start, end);
        Logger.debug(json.substring(0, 20) + "..." + json.substring(json.length() - 20)); //$NON-NLS-1$
        final JsonArdBoerseHistory histValues = JsonReader.read(json, JsonArdBoerseHistory.class);
        // check if volume data exists
        JsonArdBoerseHistory histVolumne = null;
        start = response.indexOf(BoerseArdApi.US_FDATA_HISTORY, end);
        if (start > 0) {
            start += 17;
            end = response.indexOf("});", start) + 1; //$NON-NLS-1$
            json = response.substring(start, end);
            Logger.debug(json.substring(0, 20) + "..." + json.substring(json.length() - 20)); //$NON-NLS-1$
            histVolumne = JsonReader.read(json, JsonArdBoerseHistory.class);
        }
        final WertpapierHistory historyValues = new ArdBoerseWertpapierHistory(stock.getId(), histValues, histVolumne);
        stock.setHistoryValues(historyValues);
    }

    private List<ArdStockInfoBean> search(final String searchText) throws FinanceException {

        final List<ArdStockInfoBean> results = new ArrayList<>();
        try {
            final String encodedText = URLEncoder.encode(searchText, StandardCharsets.UTF_8.name());
            final String urlString = BoerseArdApi.SEARCH_URL.replaceFirst("XXX", encodedText); //$NON-NLS-1$
            final URI uri = new URI(urlString);
            final String response = this.getResonseGETRequest(uri, true);
            Logger.debug("Response: " + response); //$NON-NLS-1$
            if (response.contains("response\":\"valid")) { //$NON-NLS-1$
                int aktPos = 0;
                while (response.indexOf(AdditionalHtmlTags.TAG_TR_BEGIN, aktPos) > 0) {
                    // ID
                    int pos = response.indexOf(AdditionalHtmlTags.TAG_A_BEGIN, aktPos);
                    final int startHref = response.indexOf("href=", pos) + 7; //$NON-NLS-1$
                    final int endHref = response.indexOf('\"', startHref);
                    String exchangeLink = response.substring(startHref, endHref);
                    if ("\\".equals(exchangeLink.substring(exchangeLink.length() - 1))) { //$NON-NLS-1$
                        exchangeLink = exchangeLink.substring(0, exchangeLink.length() - 1);
                    }
                    final int paramStart = exchangeLink.indexOf("i="); //$NON-NLS-1$
                    final String exchangeId = exchangeLink.substring(paramStart + 2);
                    // Ergebnis Text
                    int posStart = response.indexOf(AdditionalHtmlTags.TAG_CLOSE, pos);
                    int posEnd = response.indexOf(AdditionalHtmlTags.TAG_A_END, posStart);
                    final String name = StringEscapeUtils.unescapeHtml(response.substring(posStart + 1, posEnd));
                    Logger.debug("Text: " + name); //$NON-NLS-1$
                    // Ergebnis Type
                    pos = response.indexOf(AdditionalHtmlTags.TAG_TD_BEGIN, posEnd);
                    posStart = response.indexOf(AdditionalHtmlTags.TAG_CLOSE, pos);
                    posEnd = response.indexOf(AdditionalHtmlTags.TAG_TD_END, posStart);
                    final String type = StringEscapeUtils.unescapeHtml(response.substring(posStart + 1, posEnd));
                    Logger.debug("Type: " + type); //$NON-NLS-1$
                    aktPos = posEnd;
                    final ArdStockInfoBean stock = new ArdStockInfoBean(); // NOPMD
                    stock.setName(name);
                    stock.setType(type);
                    stock.setId(exchangeId);
                    results.add(stock);
                }
            } else {
                Logger.error("Invalid answer"); //$NON-NLS-1$
            }
        } catch (final URISyntaxException | UnsupportedEncodingException e) {
            Logger.error(BoerseArdApi.URL_FORMAT_ERROR, e);
            throw (FinanceException) new FinanceException("URL ist nicht korrekt formatiert: " + e.getMessage(), //$NON-NLS-1$
                    FinanceExceptionType.NETWORK).initCause(e);
        }
        return results;
    }

    /**
     * Liefert die Währung im ISO Format
     *
     * @param value
     *            StringArray aus Wert und Währung als Symbol HTML Encoded.
     * @return die Währung im ISO Format oder einen leeren String.
     */
    private String setCurrency(final String... value) {

        String currency = StringUtils.EMPTY;
        if (value.length > 1) { // NOPMD
            // Ermittle ISO code
            switch (value[1]) {
                case "&euro;": //$NON-NLS-1$
                    currency = "EUR"; //$NON-NLS-1$
                    break;
                case "&pound;": //$NON-NLS-1$
                    currency = "GBP"; //$NON-NLS-1$
                    break;
                case "$": //$NON-NLS-1$
                    currency = "USD"; //$NON-NLS-1$
                    break;
                case "&yen;": //$NON-NLS-1$
                    currency = "JPY"; //$NON-NLS-1$
                    break;
                default:
                    // Take what is there
                    currency = value[1];
                    break;
            }
        }
        Logger.debug("Currency: >" + currency + '<'); //$NON-NLS-1$
        return currency;
    }

    /**
     * @param stock
     *            der zu aktualisierende infoBean
     * @param response
     */
    private void updateValues(final ArdStockInfoBean stock, final String response) {

        Logger.trace("Response: " + response); //$NON-NLS-1$
        int start = response.indexOf("einzelkurs_header"); //$NON-NLS-1$
        start = response.indexOf("<h1>", start); //$NON-NLS-1$
        // check for headline extension
        final int extensionPos = response.indexOf("<span class=\"headline_extension\">", start); //$NON-NLS-1$
        int end = response.indexOf("</h1>", start); //$NON-NLS-1$
        if (extensionPos > 0 && extensionPos < end) {
            end = extensionPos;
        }
        final String name = response.substring(start + 4, end).trim();
        stock.setName(name);
        // aktueller Wert
        start = response.indexOf("aktueller Wert"); //$NON-NLS-1$
        int posEnd = response.indexOf("</span>", start); //$NON-NLS-1$
        String text = response.substring(start + 16, posEnd);
        text = text.replace(".", StringUtils.EMPTY); //$NON-NLS-1$
        text = text.replace(",", ".").replaceAll(AdditionalHtmlTags.NBSP, AdditionalHtmlTags.SPACE).trim(); //$NON-NLS-1$ //$NON-NLS-2$
        final String[] value = text.split(AdditionalHtmlTags.SPACE);
        // Währung
        Logger.debug(AdditionalHtmlTags.TAG_CLOSE + value[0] + "< - >"); //$NON-NLS-1$
        stock.setActValue(new BigDecimal(value[0]));
        stock.setCurrency(this.setCurrency(value));
        // Veränderung zum Vortag
        start = response.indexOf("nderung zum Vortag"); //$NON-NLS-1$
        posEnd = response.indexOf("</span>", start); //$NON-NLS-1$
        text = response.substring(start + 20, posEnd);
        text = text.replace(".", StringUtils.EMPTY); //$NON-NLS-1$
        text = text.replace(AdditionalHtmlTags.PLUSMINUS, StringUtils.EMPTY);
        text = text.replace(",", ".").replaceAll(AdditionalHtmlTags.NBSP, AdditionalHtmlTags.SPACE).trim(); //$NON-NLS-1$ //$NON-NLS-2$
        final String changeRate = text.substring(0, text.length() - 1);
        stock.setChangeRate(new BigDecimal(changeRate));
        // Datum (um sicher zu gehen per RegEx)
        posEnd = response.indexOf("Kursinformationen", start); //$NON-NLS-1$
        final String toInspect = response.substring(start, posEnd);
        final Pattern timeDate = Pattern.compile("\\d\\d.\\d\\d.\\d\\d\\d\\d&nbsp;&nbsp;\\d\\d:\\d\\d"); //$NON-NLS-1$
        final Matcher timeDateMatcher = timeDate.matcher(toInspect);
        Date date = null;
        if (timeDateMatcher.find()) {
            String strDate = timeDateMatcher.group();
            strDate = strDate.replace(AdditionalHtmlTags.NBSP, AdditionalHtmlTags.SPACE);
            Logger.debug("Date:" + strDate); //$NON-NLS-1$
            final SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy  hh:mm", Locale.GERMAN); //$NON-NLS-1$
            try {
                date = sdf.parse(strDate);
            } catch (final ParseException e) {
                Logger.error("Datum konnte nicht gelesen werden", e); //$NON-NLS-1$
            }
        }
        stock.setDate(date);
        // Börsenplatz
        start = response.indexOf("headers=\"boerse"); //$NON-NLS-1$
        start = response.indexOf(AdditionalHtmlTags.TAG_CLOSE, start) + 1;
        posEnd = response.indexOf(AdditionalHtmlTags.TAG_TD_END, start);
        final String borse = response.substring(start, posEnd);
        stock.setExchange(borse);
        // Gattung
        start = response.indexOf("headers=\"gattung"); //$NON-NLS-1$
        start = response.indexOf(AdditionalHtmlTags.TAG_CLOSE, start) + 1;
        posEnd = response.indexOf(AdditionalHtmlTags.TAG_TD_END, start);
        final String type = response.substring(start, posEnd);
        stock.setType(type);
        // Börsenplätze
        stock.setExchanges(this.getExchanges(response));
    }
}
