/*******************************************************************************
 * Copyright 2017 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europäischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie dürfen dieses Werk ausschließlich gemäß dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschränkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.jameica.json;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.math.BigDecimal;
import java.time.Instant;
import java.time.LocalDate;
import java.time.ZoneId;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;

import de.frankmuenster.jameica.json.annotations.JsonDeserialise;
import de.frankmuenster.jameica.json.annotations.JsonValueName;
import de.frankmuenster.jameica.json.exeptions.JsonCastException;
import de.willuhn.logging.Logger;

/**
 * Reader für die historischen Daten der ARD Börse
 *
 * @author Frank Münster
 */
public final class JsonReader {

    /**
     *
     */
    private static final String SETTER_PREFIX = "set"; //$NON-NLS-1$

    /**
     * Standardkonstruktor
     */
    private JsonReader() {
        // Absichtlich leer
    }

    /**
     * Liest die historische ARD Werte aus dem JSON-String
     *
     * @param json
     * @param clazz
     * @return ein JsonArdBoerseHistory Objekt.
     * @throws JsonCastException
     */
    public static <T extends Object> T read(final String json, final Class<T> clazz) throws JsonCastException {

        T returnObj = null;
        try {
            Logger.debug(json);
            returnObj = clazz.newInstance();
            final ScriptEngine engine = new ScriptEngineManager().getEngineByName("nashorn"); //$NON-NLS-1$
            final String script = "Java.asJSONCompatible(" + json + ")"; //$NON-NLS-1$ //$NON-NLS-2$
            final Object jsonObj = engine.eval(script);
            @SuppressWarnings("unchecked") // NOPMD
            final Map<String, Object> jsonMap = (Map<String, Object>) jsonObj;
            returnObj = JsonReader.readValues(jsonMap, returnObj);
        } catch (final ScriptException | ClassCastException | NoSuchMethodException | SecurityException
                | IllegalAccessException | IllegalArgumentException | InvocationTargetException | InstantiationException
                | ClassNotFoundException e) {
            Logger.error(e.getLocalizedMessage());
            throw new JsonCastException("Fehler beim Einlesen des JSON Strings.", e); //$NON-NLS-1$
        }
        return returnObj;
    }

    private static <T extends Object> T readValues(final Map<String, Object> jsonMap, final T returnObj)
            throws NoSuchMethodException, IllegalAccessException, InvocationTargetException, InstantiationException,
            ClassNotFoundException {

        for (final Field field : returnObj.getClass().getDeclaredFields()) {
            if (field.isAnnotationPresent(JsonValueName.class)) {

                final Object value = JsonReader.inspectField(field, jsonMap);

                if (value != null) {
                    Method getter = null;
                    final String name = Character.toUpperCase(field.getName().charAt(0)) + field.getName().substring(1);

                    if (value instanceof List) {
                        getter = returnObj.getClass().getDeclaredMethod(JsonReader.SETTER_PREFIX + name, List.class);
                    } else {
                        getter = returnObj.getClass().getDeclaredMethod(JsonReader.SETTER_PREFIX + name,
                                value.getClass());
                    }
                    getter.invoke(returnObj, value);
                }
            }
        }
        return returnObj;
    }

    /**
     * Inspects a field and checks if it is primitiv value or a list.
     *
     * @param field
     * @param jsonMap
     * @return the value of field of found
     * @throws ClassNotFoundException
     * @throws InstantiationException
     * @throws InvocationTargetException
     * @throws IllegalAccessException
     * @throws NoSuchMethodException
     */
    private static <U> Object inspectField(final Field field, final Map<String, Object> jsonMap)
            throws NoSuchMethodException, IllegalAccessException, InvocationTargetException, InstantiationException,
            ClassNotFoundException {

        Object value = null;

        final JsonValueName annotation = field.getAnnotation(JsonValueName.class);
        final Class<?> type = field.getType();

        if (field.isAnnotationPresent(JsonDeserialise.class)) {
            // Gibt es eine eigenen Deserialiser
            final JsonDeserialise annotation1 = field.getAnnotation(JsonDeserialise.class);
            final String clazz = annotation1.value();
            final Deserialiser<?> deserialiser = (Deserialiser<?>) Class.forName(clazz).newInstance();
            value = deserialiser.deserialize(jsonMap.get(annotation.value()));
        } else if (type.equals(List.class)) {
            // Ist es eine Liste
            final ParameterizedType type1 = (ParameterizedType) field.getGenericType();
            final Type listType = type1.getActualTypeArguments()[0];
            final Class<?> clazz = Class.forName(listType.getTypeName());
            final List<U> listValues = JsonReader.createListOfType();
            @SuppressWarnings("unchecked")
            final List<Object> dataList = (List<Object>) jsonMap.get(annotation.value());
            if (dataList != null) {
                for (final Object data : dataList) {
                    Logger.debug(data.getClass().toGenericString());
                    @SuppressWarnings("unchecked")
                    final Map<String, Object> jsonMap1 = (Map<String, Object>) data;
                    Logger.debug("Map size: " + jsonMap1.size()); //$NON-NLS-1$
                    @SuppressWarnings("unchecked")
                    U listElement = (U) clazz.newInstance();
                    listElement = JsonReader.readValues(jsonMap1, listElement);
                    listValues.add(listElement);
                }
            }
            value = listValues;
        } else {
            // normaler Value
            value = JsonReader.readValue(jsonMap, annotation.value(), type);
        }

        return value;

    }

    private static <T extends Object> T readValue(final Map<String, Object> jsonMap, final String valueName,
            final Class<T> clazz) {

        T returnValue = null;
        final Object jsonValue = jsonMap.get(valueName);
        if (jsonValue instanceof Double) {
            if (clazz.equals(LocalDate.class)) {
                returnValue = clazz.cast(JsonReader.doubleToLocalDate((Double) jsonValue));
            } else if (clazz.equals(BigDecimal.class)) {
                returnValue = clazz.cast(BigDecimal.valueOf((Double) jsonValue));
            }
        }
        return returnValue;
    }

    private static LocalDate doubleToLocalDate(final Double date) {

        final Long dateLong = date.longValue();
        return Instant.ofEpochMilli(dateLong).atZone(ZoneId.systemDefault()).toLocalDate();
    }

    private static <T> List<T> createListOfType() {

        return new ArrayList<>();
    }
}
